@@ -1,11 +1,14 @@
+from collections import namedtuple
+
 try:
     from collections import UserDict
 except ImportError:
     from UserDict import UserDict
 
-from django.utils.encoding import StrAndUnicode
-from django.template.loader import render_to_string
 from django.db.models.fields import FieldDoesNotExist
+from django.template.loader import render_to_string
+from django.utils.encoding import StrAndUnicode
+from django.forms.util import flatatt
 
 # Sane boundary constants
 MINIMUM_PAGE_LENGTH = 5
@@ -13,18 +16,14 @@ MINIMUM_PAGE_LENGTH = 5
 DEFAULT_OPTIONS = {
     'columns': [],  # table headers
     'ordering': [],  # override to Model._meta.ordering
-    'filters': {},  # field_name__lookuptype: value
     'start_offset': 0,  # results to skip ahead
     'page_length': 25,  # length of a single result page
-    'search': None,  # client search string
+    'search': '',  # client search string
     'search_fields': [],  # extra ORM paths to search; not displayed
     'unsortable_columns': [],  # table headers not allowed to be sorted
     'hidden_columns': [],  # table headers to be generated, but hidden by the client
     'structure_template': "datatableview/default_structure.html",
     'result_counter_id': 'id_count',  # HTML element ID to display the total results
-
-    # TODO: Support additional field options:
-    # 'exclude': [],
 }
 
 # Since it's rather painful to deal with the datatables.js naming scheme in Python, this map changes
@@ -69,7 +68,9 @@ _javascript_boolean = {
     True: 'true',
     False: 'false',
 }
-
+FieldDefinitionTuple = namedtuple('FieldDefinitionTuple', ['pretty_name', 'fields', 'callback'])
+ColumnOrderingTuple = namedtuple('ColumnOrderingTuple', ['order', 'column_index', 'direction'])
+ColumnInfoTuple = namedtuple('ColumnInfoTuple', ['pretty_name', 'attrs'])
 
 def resolve_orm_path(model, orm_path):
     """
@@ -107,6 +108,39 @@ def get_model_at_related_field(model, attr):
     return model
 
 
+def get_first_orm_bit(field_definition):
+    """ Returns the first ORM path component of a field definition's declared db field. """
+    column = get_field_definition(field_definition)
+
+    if not column.fields:
+        return None
+
+    return column.fields[0].split('__')[0]
+
+
+def get_field_definition(field_definition):
+    """ Normalizes a field definition into its component parts, even if some are missing. """
+    if not isinstance(field_definition, (tuple, list)):
+        field_definition = [field_definition]
+    else:
+        field_definition = list(field_definition)
+
+    if len(field_definition) == 1:
+        field = [None, field_definition, None]
+    elif len(field_definition) == 2:
+        field = field_definition + [None]
+    elif len(field_definition) == 3:
+        field = field_definition
+    else:
+        raise ValueError("Invalid field definition format.")
+
+    if not isinstance(field[1], (tuple, list)):
+        field[1] = (field[1],)
+    field[1] = tuple(name for name in field[1] if name is not None)
+
+    return FieldDefinitionTuple(*field)
+
+
 class DatatableStructure(StrAndUnicode):
     """
     A class designed to be echoed directly to into template HTML to represent a skeleton table
@@ -114,20 +148,20 @@ class DatatableStructure(StrAndUnicode):
 
     """
 
-    def __init__(self, ajax_url, model, options):
+    def __init__(self, ajax_url, options, model=None):
         self.url = ajax_url
-        self.model = model
         self.options = options
+        self.model = model
 
-        ordering = options.ordering or model._meta.ordering
         self.ordering = {}
-        for i, name in enumerate(ordering):
-            plain_name = name.lstrip('-+')
-            index = options.get_column_index(plain_name)
-            if index == -1:
-                continue
-            sort_direction = 'desc' if name[0] == '-' else 'asc'
-            self.ordering[plain_name] = (i, index, sort_direction)
+        if options.ordering:
+            for i, name in enumerate(options.ordering):
+                plain_name = name.lstrip('-+')
+                index = options.get_column_index(plain_name)
+                if index == -1:
+                    continue
+                sort_direction = 'desc' if name[0] == '-' else 'asc'
+                self.ordering[plain_name] = ColumnOrderingTuple(i, index, sort_direction)
 
     def __unicode__(self):
         return render_to_string(self.options.structure_template, {
@@ -151,29 +185,31 @@ class DatatableStructure(StrAndUnicode):
         """
         Returns an iterable of 2-tuples in the form
 
-            ("column_name", ' data-bSortable="true"',)
+            ("Pretty name", ' data-bSortable="true"',)
 
         """
 
         column_info = []
-        model_fields = self.model._meta.get_all_field_names()
+        if self.model:
+            model_fields = self.model._meta.get_all_field_names()
+        else:
+            model_fields = []
 
-        for name in self.options.columns:
-            if isinstance(name, (tuple, list)):
-                # Take the friendly representation
-                name = pretty_name = name[0]
-            elif name in model_fields:
-                # Get the raw field name's verbose_name attribute.
-                field, model, direct, m2m = self.model._meta.get_field_by_name(name)
-                pretty_name = field.verbose_name.capitalize()
+        for column in self.options.columns:
+            column = get_field_definition(column)
+            pretty_name = column.pretty_name
+            column_name = column.pretty_name
+            if column.fields and column.fields[0] in model_fields:
+                ordering_name = column.fields[0]
+                if not pretty_name:
+                    field = self.model._meta.get_field_by_name(column.fields[0])[0]
+                    column_name = field.name
+                    pretty_name = field.verbose_name
             else:
-                # Purely virtual column name
-                pretty_name = name.capitalize()
+                ordering_name = pretty_name
 
-            attributes = self.get_column_attributes(name)
-
-            attributes_string = ' '.join('{0}="{1}"'.format(*item) for item in attributes.items())
-            column_info.append((pretty_name, attributes_string))
+            attributes = self.get_column_attributes(ordering_name)
+            column_info.append(ColumnInfoTuple(pretty_name, flatatt(attributes)))
 
         return column_info
 
@@ -190,20 +226,7 @@ class DatatableStructure(StrAndUnicode):
 
 
 class DatatableOptions(UserDict):
-    """
-    ``columns``: An iterable of column names.  If any item is a 2-tuple, it is treated as a
-    description in the form of ('Display Name', 'attribute_name'), where 'attribute_name' will be
-    looked up in the following order: instance.attribute_name(), instance.attribute_name,
-    view.attribute_name(instance)
-
-    ``ordering``: A list or tuple of column names to sort by.  If empty or ``None``, the model's
-    default Meta.ordering option is respected.  Names corresponding to a database-backed field can
-    be dealt with in the database, but virtual columns that exist as compound data cells need to be
-    handled in code, which has a hard efficiency limit.  Mixing real and virtual columns tries to
-    be as efficient as possible by letting the database do the sorting first, but ultimately
-    triggers the code-driven ordering.
-
-    """
+    """ Normalizes all options to values that are guaranteed safe. """
 
     def __init__(self, model, query_parameters, *args, **kwargs):
         self._model = model
@@ -211,7 +234,7 @@ class DatatableOptions(UserDict):
         # Core options, not modifiable by client updates
         if 'columns' not in kwargs:
             model_fields = model._meta.local_fields
-            kwargs['columns'] = map(lambda f: f.verbose_name.capitalize(), model_fields)
+            kwargs['columns'] = map(lambda f: (f.verbose_name, f.name), model_fields)
 
         if 'hidden_columns' not in kwargs or kwargs['hidden_columns'] is None:
             kwargs['hidden_columns'] = []
@@ -219,20 +242,21 @@ class DatatableOptions(UserDict):
         if 'search_fields' not in kwargs or kwargs['search_fields'] is None:
             kwargs['search_fields'] = []
 
+        if 'unsortable_columns' not in kwargs or kwargs['unsortable_columns'] is None:
+            kwargs['unsortable_columns'] = []
+
         # Absorb query GET params
         kwargs = self._normalize_options(query_parameters, kwargs)
 
         UserDict.__init__(self, DEFAULT_OPTIONS, *args, **kwargs)
 
         self._flat_column_names = []
-        for field_name in self.columns:
-            if isinstance(field_name, (tuple, list)):
-                pretty_name, field_name = field_name[:2]
-
-            if not field_name or isinstance(field_name, (tuple, list)):
-                field_name = pretty_name
-
-            self._flat_column_names.append(field_name)
+        for column in self.columns:
+            column = get_field_definition(column)
+            flat_name = column.pretty_name
+            if column.fields:
+                flat_name = column.fields[0]
+            self._flat_column_names.append(flat_name)
 
     def __getattr__(self, k):
         try:
@@ -294,24 +318,20 @@ class DatatableOptions(UserDict):
                     if column_index >= len(options['columns']):
                         continue
 
-                    field_name = options['columns'][column_index]
-                    if isinstance(field_name, (tuple, list)):
-                        name, field_name = field_name[:2]
+                    column = options['columns'][column_index]
+                    column = get_field_definition(column)
+                    is_local_field = column.fields[0] in self._model._meta.get_all_field_names()
+                    if not column.fields or len(column.fields) > 1 or not is_local_field:
+                        field_name = '!{0}'.format(column_index)
 
-                        # If the database source for the field is None, then this column will be
-                        # forcefully sorted in code.  If the field_name is an iterable of compound
-                        # sources, the final output from the data method should also be used.
-                        if not field_name or isinstance(field_name, (tuple, list)):
-                            field_name = '!{0}'.format(column_index)
+                    if is_local_field:
+                        name = column.fields[0]
+                        field_name = name
                     else:
-                        name = field_name
-
-                        # If the singular column name isn't a model field, mark it for manual handling
-                        if field_name not in self._model._meta.get_all_field_names():
-                            field_name = '!{0}'.format(column_index)
+                        name = column.pretty_name
 
                     # Reject requests for unsortable columns
-                    if name in options.get('unsortable_columns', []):
+                    if name in options['unsortable_columns']:
                         continue
 
                     # Get the client's requested sort direction
@@ -326,6 +346,8 @@ class DatatableOptions(UserDict):
                         continue
 
                     options['ordering'].append('%s%s' % (sort_modifier, field_name))
+        if not options['ordering'] and self._model:
+            options['ordering'] = self._model._meta.ordering
 
         return options
 
@@ -343,7 +365,7 @@ class ObjectListResult(list):
     _dtv_unpaged_total = None
 
 
-def get_datatable_structure(ajax_url, model, options):
+def get_datatable_structure(ajax_url, options, model=None):
     """
     Uses ``options``, a dict or DatatableOptions, into a ``DatatableStructure`` for template use.
 
@@ -352,26 +374,20 @@ def get_datatable_structure(ajax_url, model, options):
     if not isinstance(options, DatatableOptions):
         options = DatatableOptions(model, {}, **options)
 
-    return DatatableStructure(ajax_url, model, options)
+    return DatatableStructure(ajax_url, options, model=model)
 
 
-def split_real_fields(model, field_list, key=None):
+def split_real_fields(model, field_list):
     """
     Splits a list of field names on the first name that isn't in the model's concrete fields.  This
     is used repeatedly for allowing a client to request sorting or filtering on virtual or compound
     columns in the display.
 
-    If ``key`` is specified, it is used to access items in ``field_list`` for the comparison, in
-    the same fashion as the built-in ``sort`` function.
-
     Returns a 2-tuple, where the database can safely handle the first item, and the second must be
     handled in code.
 
     """
 
-    if key:
-        field_list = map(key, field_list)
-
     i = 0
 
     for i, field_name in enumerate(field_list):
@@ -397,8 +413,8 @@ def filter_real_fields(model, fields, key=None):
 
     """
 
-    field_map = dict(zip(map(key, fields), fields))
-
+    field_hints = zip(map(key, fields), fields)
+    field_map = dict(field_hints)
     field_list = set(field_map.keys())
     concrete_names = set(model._meta.get_all_field_names())
 
@@ -407,6 +423,12 @@ def filter_real_fields(model, fields, key=None):
     virtual_fields = field_list.difference(concrete_names)
 
     # Get back the original data items that correspond to the found data
-    db_fields = filter(lambda f: (key(f) if key else f) in concrete_fields, fields)
-    virtual_fields = map(field_map.get, virtual_fields)
+    db_fields = []
+    virtual_fields = []
+    for bit, field in field_hints:
+        if bit in concrete_fields:
+            db_fields.append(field)
+        else:
+            virtual_fields.append(field)
     return db_fields, virtual_fields
+
